-- Genetics Library version 1.0 by Jeffrey Fielding (JJProg@cyberbury.net)
-- Please send me any questions, comments, or suggestions you have.
include get.e
-- Random number routines
atom IX
constant IA = 16807, I15 = 32768, I16 = 65536, IP = 2147483647
constant DK = 1/IP
global function random()
    atom IXH, IXA, LFTL, IFH, K
    IXH = floor(IX/I16)
    IXA = (IX-IXH*I16)*IA
    LFTL = floor(IXA/I16)
    IFH = IXH *IA + LFTL
    K = floor(IFH/I15)
    IX = (((IXA - LFTL * I16) -IP) + (IFH - K * I15) * I16) + K
    if IX < 0 then
	IX += IP
    end if
    return IX*DK
end function
IX = 2000000000
function randomf(object x)
    if atom(x) then
	return random()
    else
	for i = 1 to length(x) do
	    x[i] = randomf(x[i])
	end for
	return x
    end if
end function
-- End of random number routines
constant LEARN_SIZE = 40   -- The number of "lifeforms"
constant LEARN_RATE = 0.01 -- The maximum absolute mutation
sequence data              -- All the "lifeforms" and their genes and fitness
integer currentNode        -- The node or "lifeform" that is currently selected
-- loadGeneticData(fileName)
-- Loads the genetic data from fileName into memory
-- If it can't, it creates all blank lifeforms
global procedure loadGeneticData(sequence fileName)
    integer f
    f = open(fileName,"r")
    if f = -1 then
	data = repeat({{}},LEARN_SIZE)
	return
    end if
    data = get(f)
    if data[1] != GET_SUCCESS then
	data = repeat({{}},LEARN_SIZE)
	return
    end if
    data = data[2]
    close(f)
end procedure
-- writeGeneticData(fileName)
-- Writes the genetic data to fileName so it can be reloaded later
-- by loadGeneticData(fileName)
global procedure writeGeneticData(sequence fileName)
    integer f
    f = open(fileName,"w")
    print(f,data)
    close(f)
end procedure
-- selectNode()
-- selects a node or "lifeform" to use
global procedure selectNode()
    currentNode = rand(LEARN_SIZE)
end procedure
-- setNode(node)
-- sets a node or "lifeform" to use
global procedure setNode(integer node)
    currentNode = node
end procedure
-- mutate()
-- Replaces the current node with a mutated form of a randomly selected
-- "parent"
global procedure mutate()
    data[currentNode] = data[rand(LEARN_SIZE)]
    for i = 1 to length(data[currentNode][1]) do
	data[currentNode][1][i][2] += (random()*(2*LEARN_RATE))-LEARN_RATE
    end for
end procedure
-- naturallySelect()
-- Selects a "lifeform" to mutate based on its fitness
-- lower fitness, more likely to be replaced by a mutated "lifeform"
-- If a node hasn't been assigned a fitness, it will be the average
-- fitness, or just a random value from
-- 0 to 1 if none of the lifeforms have been assigned a fitness.
global procedure naturallySelect()
    atom average
    atom rnd, min, max, sum
    integer n
    sequence s
    average = 0
    n = 0
    for i = 1 to length(data) do
	if length(data[i]) = 2 then
	    average += data[i][2]
	    n += 1
	end if
    end for
    if n then
	average /= n
    end if
    for i = 1 to length(data) do
	if length(data[i]) = 1 then
	    if n then
		data[i] &= average
	    else
		data[i] &= random()
	    end if
	end if
	if i = 1 then
	    min = data[i][2]
	    max = data[i][2]
	else
	    if min > data[i][2] then
		min = data[i][2]
	    elsif max < data[i][2] then
		max = data[i][2]
	    end if
	end if
    end for
    s = repeat(1,length(data))
    if max != min then
	sum = 0
	for i = 1 to length(data) do
	    s[i] = (data[i][2]-min)/(max-min)
	    sum += s[i]
	end for
    else
	sum = length(data)
    end if
    rnd = random() * sum
    sum = 0
    for i = 1 to length(data) do
	sum += s[i]
	if sum >= rnd then
	    n = currentNode
	    currentNode = i
	    mutate()
	    currentNode = n
	    exit
	end if
    end for
end procedure
-- value = getGene(state)
-- Gets the value of the gene referenced by state
-- If no gene has that state, it returns -1
global function getGene(object state)
    integer index
    index = 0
    for i = 1 to length(data[currentNode][1]) do
	trace(1)
	if compare(data[currentNode][1][i][1],state) = 0 then
	    index = i
	    exit
	end if
    end for
    if not index then
	return -1
    else
	return data[currentNode][1][index][2]
    end if
end function
-- createGene(state,format)
-- Creates a gene referenced by state, with the value in the format spesified
-- by format. For example: createGene(1,{0,{0,0}}) would create a gene
-- with state 1 and gene value {random(),{random(),random()}} where random()
-- generates a random value from 0 to 1
global procedure createGene(object state, object format)
    data[currentNode][1] = append(data[currentNode][1],{state,randomf(format)})
end procedure
-- setFitness(fitness)
-- fitness is an atom, sets fitness of current "lifeform" to fitness
global procedure setFitness(atom fitness)
    if length(data[currentNode]) = 1 then
	data[currentNode] &= fitness
    else
	data[currentNode][2] = fitness
    end if
end procedure
-- fitness = getFitness()
-- returns the fitness of the current "lifeform" or -1 if there is none
global function getFitness()
    if length(data[currentNode]) = 1 then
	return -1
    else
	return data[currentNode][2]
    end if
end function
-- Set and get the data
global procedure setGeneticData(sequence d)
    data = d
end procedure
global function getGeneticData()
    return data
end function
